from datetime import datetime
import warnings
from subprocess import Popen, PIPE
import re
import GRID_LRT.auth.grid_credentials as grid_creds
from GRID_LRT.auth.get_picas_credentials import picas_cred
from GRID_LRT import Token
from GRID_LRT.Staging.srmlist import srmlist
import humanfriendly

import pdb
class GSIFile(object):
    def __init__(self, location, parent_dir=None):
        _ = grid_creds.grid_credentials_enabled()
        self.location = location
        self.protocol = location.split("://")[0]
        self.port = self._get_port(location)

        if parent_dir:
            self._build_from_parent_dir(parent_dir)
        else:
            self._subfiles = None
            self.is_dir, self.parent_dir = self._check_if_directory(location)
            self._internal = self._test_file_exists(location)
        self.datetime = self._extract_date(self._internal) 
        self.size = self._get_size()
        if self.is_dir:
            self.is_file = False
            self.filename = self.location.split('/')[-1]
        else:
            self.is_file = True
            self.filename = self._internal[-1]

    def _build_from_parent_dir(self, parent_dir):
        self.parent_dir = parent_dir.location
        self._internal = self._find_item_in_uberftp_result(parent_dir._subfiles)
        if self._internal[0][0]=='d':
            self.is_dir = True
        elif self._internal[0][0] == '-':
            self.is_dir = False

    def _get_size(self):
        self._bytesize = self._internal[4]
        human_size = humanfriendly.parse_size(self._bytesize)
        return humanfriendly.format_size(human_size)

    def get_dir_size(self):
        """Will go one level deeper and sum the sizes of all objects. If the objects are directories, you'll
        have to do the iteration yourself!"""
        total = humanfriendly.parse_size(self.size)
        for f in self.list_dir():
            total += humanfriendly.parse_size(f.size)
        print(total)
        t = humanfriendly.parse_size(str(total))
        return  humanfriendly.format_size(t)

    def _check_if_directory(self,location):
        self.num_subdir = len([i for i in location.split('gsiftp://')[1].split('/') if i]) - 1
        filename = location.split('/')[-1]
        parent_dir = self.get_parent_dir()
        parent_dir = parent_dir.replace(self.protocol+':/',self.protocol+"://")  #TODO Make this cleaner (new staticmethod)
        res, err = self._uberftpls(parent_dir)
        if '550' in err:
            warnings.warn('Parent directory inaccessible!: {}'.format(err))
            return None, parent_dir #Doesn't know if it's a file or dir if can't get parent_dir
        status = [x for x in res.split('\r\n') if filename in x]
        if status[0][0]=='d':
            return True, parent_dir
        else:
            return False, parent_dir

    def __repr__(self):
        return "<GRID_LRT.Storage.gsifile.GSIFile {} {} located in {} >".format("File" if self.is_file else "Folder", 
                self.filename, self.location)

    def _test_file_exists(self,location):
        result, error = self._uberftpls(location)
        if result=="" and error=="":
            if location[-1]=='/':
                location = location[:-1]
                result, error = self._uberftpls("/".join(location.split('/')[:-1]))
        if "No match " in error:
            raise Exception("file %s cannot be found: %s"%(location, error))
        if self.is_dir:
            result,err = self._uberftpls(self.parent_dir)
            result = self._find_item_in_uberftp_result(result) 
        return result

    def _find_item_in_uberftp_result(self,result):
        for i in result.split('\r\n'):
            if self.location[-1]=='/':
                location = self.location[:-1]
            if self.location.split('/')[-1] in i:
                result = i
                break
        result = result.split()
        return result

    @staticmethod
    def _extract_date(data):
        if data[-2] not in ['2018','2017','2016','2015']:
            date = data[-4]+" "+data[-3]+" " + str(datetime.now().year)
            time = data[-2]
        else:
            date = data[-4]+" "+data[-3]+" "+data[-2]
            time = "00:00"
        file_datetime = datetime.strptime(date+"-"+time, "%b %d %Y-%H:%M")
        return file_datetime

    @staticmethod
    def _get_port(location):
        try:
            port = re.search(':[0-9]{4}/',location).group(0)[1:][:-1]
            return port
        except AttributeError:
            return None

    def _donotdelete(self,location):
        """Raises Exception if you try to delete files or folders whose parent is
        one of these folders"""
        locations = ['gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/archive',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/spectroscopy-migrated',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/sksp_natalie',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/distrib',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/sandbox',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/LGPPP',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/PiLL',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/SKSP',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/test',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/NDPPP'
#                     'gsiftp://gridftp.grid.sara.n:2811l/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/DISC'
                     ]
        if self.parent_dir in locations:
            raise Exception("FORBIDDEN to delete this {} because its parent is {}".format(
                            'file' if self.is_file else 'folder', self.parent_dir))

    def delete(self):
        parent_dir = self.get_parent_dir()
        self._donotdelete(parent_dir)
        if self.is_dir and not self.is_empty():
            raise Exception("Not allowed to delete a folder that isn't empty yet" )
        del_proc = Popen(['uberftp','-rm',self.location], stdout=PIPE, stderr=PIPE)
        res, err = del_proc.communicate()
        if not err:
            return
        else: 
            warnings.warn('Deleting failed: {}'.format(err))
    
    def get_parent_dir(self):
        location = "/".join([i for i in self.location.split('/') if i])
        parent_dir = "/".join(location.split('/')[:-1])
        return parent_dir

    def _get_num_files(self):
        """If self is folder, gets the number of files in it"""
        if self.is_file:
            return 1
        res, err = self._uberftpls(self.location)
        num_files = len([x for x in res.split('\r\n') if x])
        return num_files

    def _uberftpls(self, location):
        sub = Popen(['uberftp','-ls', location], stdout=PIPE, stderr=PIPE)
        res, err = sub.communicate()
        if err:
            warnings.warn("Uberftp -ls gave us an error for location {}: {}".format(location, err))
        return res, err

    def is_empty(self):
        if self._get_num_files() == 0:
            return True
        return False

    def list_dir(self):
        """Checks the status of all srm links in the folder given. 
        Returns a list of localities
        """     
        if self.is_dir:
           results, error= self._uberftpls(self.location)
        else: 
            return []
        if results == '':
            return []
        results = results.strip().split("\r\n")
        file_locs = [self.location +"/"+str(i.split()[-1])
                for i in results if i]
        self._subfiles, _= self._uberftpls(self.location)
        files_list = [GSIFile(i, parent_dir=self) for i,r in zip(file_locs,results)]
        return files_list


class MockGSIFile(GSIFile):
    def __init__(self, location):
        """uses text from uberftp -ls to initialize, saved in a file"""
        d = open(location,'r').read()
        self._filedata = '\r\n'.join(d.split('\r\n')[1:])
        self.location =d.split('\r\n')[0]
        self.parent_dir = self.location.split('/')[-1]
        self._internal = self._find_item_in_uberftp_result(self._filedata)
        if len(d)>2:
            self.is_dir = True
            self.is_file = False
        else:
            self.is_dir = False
            self.is_file = True

    def list_dir(self):
        results = self._filedata.split('\r\n')[1:]
        file_locs = [self.location +"/"+str(i.split()[-1])
                                for i in results if i]
        return file_locs





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Load design documents from the filesystem into a dict.
Subset of couchdbkit/couchapp functionality.

Description
-----------

Convert a target directory into an object (dict).

Each filename (without extension) or subdirectory name is a key in this object.

For files, the utf-8-decoded contents are the value, except for .json files
which are first decoded as json.

Subdirectories are converted into objects using the same procedure and then
added to the parent object.

Typically used for design documents. This directory tree::

  .
    ├── filters
    │   └── forms_only.js
    ├── _id
    ├── language
    ├── lib
    │   └── validate.js
    └── views
        ├── view_a
        │   └── map.js
        ├── view_b
        │   └── map.js
        └── view_c
            └── map.js

Becomes this object::

    {
      "views": {
        "view_a": {
          "map": "function(doc) { ... }"
        },
        "view_b": {
          "map": "function(doc) { ... }"
        },
        "view_c": {
          "map": "function(doc) { ... }"
        }
      },
      "_id": "_design/name_of_design_document",
      "filters": {
        "forms_only": "function(doc, req) { ... }"
      },
      "language": "javascript",
      "lib": {
        "validate": "// A library for validations ..."
      }
    }

"""

from __future__ import unicode_literals, absolute_import

import os.path
import pprint
import codecs
import json

class DuplicateKeyError(ValueError):
    pass

def load_design_doc(directory, strip=False, predicate=lambda x: True):
    """
    Load a design document from the filesystem.

    strip: remove leading and trailing whitespace from file contents,
        like couchdbkit.

    predicate: function that is passed the full path to each file or directory.
        Each entry is only added to the document if predicate returns True.
        Can be used to ignore backup files etc.
    """
    objects = {}

    if not os.path.isdir(directory):
        raise OSError("No directory: '{0}'".format(directory))

    for (dirpath, dirnames, filenames) in os.walk(directory, topdown=False):
        key = os.path.split(dirpath)[-1]
        ob = {}
        objects[dirpath] = (key, ob)

        for name in filenames:
            fkey = os.path.splitext(name)[0]
            fullname = os.path.join(dirpath, name)
            if not predicate(fullname): continue
            if fkey in ob:
                raise DuplicateKeyError("file '{0}' clobbers key '{1}'"
                                        .format(fullname, fkey))
            with codecs.open(fullname, 'r', 'utf-8') as f:
                contents = f.read()
                if name.endswith('.json'):
                    contents = json.loads(contents)
                elif strip:
                    contents = contents.strip()
                ob[fkey] = contents

        for name in dirnames:
            if name == '_attachments':
                raise NotImplementedError("_attachments are not supported")
            fullpath = os.path.join(dirpath, name)
            if not predicate(fullpath): continue
            subkey, subthing = objects[fullpath]
            if subkey in ob:
                raise DuplicateKeyError("directory '{0}' clobbers key '{1}'"
                                        .format(fullpath,subkey))
            ob[subkey] = subthing

    return ob


def main():
    import sys
    try:
        directory = sys.argv[1]
    except IndexError:
        sys.stderr.write("Usage:\n\t{0} [directory]\n".format(sys.argv[0]))
        sys.exit(1)
    obj = load_design_doc(directory)
    sys.stdout.write(json.dumps(obj, indent=2))


if __name__ == "__main__":
    main()



"""GRID_LRT: Grid LOFAR Tools"""
from subprocess import PIPE, Popen
import os
import socket

__all__ = ["storage", 'auth', "application", "Staging", 'sandbox', 'Token']
__version__ = "0.5.0rc3"
__author__ = "Alexandar P. Mechev"
__copyright__ = "2018 Alexandar P. Mechev"
__credits__ = ["Alexandar P. Mechev", "Natalie Danezi", "J.B.R. Oonk"]
__bibtex__ = """@misc{apmechev:2018,
      author       = {Alexandar P. Mechev} 
      title        = {apmechev/GRID_LRT: v0.4.0},
      month        = aug,
      year         = 2018,
      doi          = {10.5281/zenodo.1341127},
      url          = {https://doi.org/10.5281/zenodo.1341127}
    }"""
__license__ = "GPL 3.0"
__maintainer__ = "Alexandar P. Mechev"
__email__ = "LOFAR@apmechev.com"
__status__ = "Production"
__date__ = "2018-09-29"



def get_git_hash():
    """Gets the git hash using git describe"""
    g_hash = ""
    proc = Popen(["git", "describe"], stdout=PIPE, stderr=PIPE)
    label = proc.communicate()[0].strip().decode("utf-8") 
    if label:
        g_hash = label[0]
        githashfile = __file__.split('__init__')[0]+"__githash__"
        if os.path.exists(githashfile):
            with open(githashfile) as _file:
                file_hash = _file.read()
        else:
            file_hash = ""
        if __version__ in g_hash:
            g_hash = g_hash.split(__version__)[1]
        if g_hash not in file_hash:
            with open(githashfile, 'w') as _file:
                _file.write(str(g_hash))
    return g_hash


#if socket.gethostname() != 'loui.grid.surfsara.nl':
#    RuntimeWarning("You're not running on loui. Some features will not work!")

__commit__ = get_git_hash()




import sys

if sys.version_info[0] < 3:
    from GRID_LRT.couchdb.util2 import *
else:
    from GRID_LRT.couchdb.util3 import *

def pyexec(code, gns, lns):
    # http://bugs.python.org/issue21591
    exec(code, gns, lns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2009 Christopher Lenz
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

"""Simple HTTP client implementation based on the ``httplib`` module in the
standard library.
"""

from base64 import b64encode
from datetime import datetime
import errno
import socket
import time
import sys
import ssl

try:
    from threading import Lock
except ImportError:
    from dummy_threading import Lock

try:
    from http.client import BadStatusLine, HTTPConnection, HTTPSConnection
except ImportError:
    from httplib import BadStatusLine, HTTPConnection, HTTPSConnection

try:
    from email.Utils import parsedate
except ImportError:
    from email.utils import parsedate

from GRID_LRT.couchdb import json
from GRID_LRT.couchdb import util

__all__ = ['HTTPError', 'PreconditionFailed', 'ResourceNotFound',
           'ResourceConflict', 'ServerError', 'Unauthorized', 'Forbidden',
           'RedirectLimit', 'Session', 'Resource']
__docformat__ = 'restructuredtext en'


if sys.version < '2.7':

    from httplib import CannotSendHeader, _CS_REQ_STARTED, _CS_REQ_SENT

    class NagleMixin:
        """
        Mixin to upgrade httplib connection types so headers and body can be
        sent at the same time to avoid triggering Nagle's algorithm.

        Based on code originally copied from Python 2.7's httplib module.
        """

        def endheaders(self, message_body=None):
            if self.__dict__['_HTTPConnection__state'] == _CS_REQ_STARTED:
                self.__dict__['_HTTPConnection__state'] = _CS_REQ_SENT
            else:
                raise CannotSendHeader()
            self._send_output(message_body)

        def _send_output(self, message_body=None):
            self._buffer.extend(("", ""))
            msg = "\r\n".join(self._buffer)
            del self._buffer[:]
            if isinstance(message_body, str):
                msg += message_body
                message_body = None
            self.send(msg)
            if message_body is not None:
                self.send(message_body)

    class HTTPConnection(NagleMixin, HTTPConnection):
        pass

    class HTTPSConnection(NagleMixin, HTTPSConnection):
        pass


class HTTPError(Exception):
    """Base class for errors based on HTTP status codes >= 400."""


class PreconditionFailed(HTTPError):
    """Exception raised when a 412 HTTP error is received in response to a
    request.
    """


class ResourceNotFound(HTTPError):
    """Exception raised when a 404 HTTP error is received in response to a
    request.
    """


class ResourceConflict(HTTPError):
    """Exception raised when a 409 HTTP error is received in response to a
    request.
    """


class ServerError(HTTPError):
    """Exception raised when an unexpected HTTP error is received in response
    to a request.
    """


class Unauthorized(HTTPError):
    """Exception raised when the server requires authentication credentials
    but either none are provided, or they are incorrect.
    """


class Forbidden(HTTPError):
    """Exception raised when the request requires an authorisation that the
    current user does not have.
    """


class RedirectLimit(Exception):
    """Exception raised when a request is redirected more often than allowed
    by the maximum number of redirections.
    """


CHUNK_SIZE = 1024 * 8

class ResponseBody(object):

    def __init__(self, resp, conn_pool, url, conn):
        self.resp = resp
        self.chunked = self.resp.msg.get('transfer-encoding') == 'chunked'
        self.conn_pool = conn_pool
        self.url = url
        self.conn = conn

    def __del__(self):
        if not self.chunked:
            self.close()
        else:
            self.resp.close()
            if self.conn:
                # Since chunked responses can be infinite (i.e. for
                # feed=continuous), and we want to avoid leaking sockets
                # (even if just to prevent ResourceWarnings when running
                # the test suite on Python 3), we'll close this connection
                # eagerly. We can't get it into the clean state required to
                # put it back into the ConnectionPool (since we don't know
                # when it ends and we can only do blocking reads). Finding
                # out whether it might in fact end would be relatively onerous
                # and require a layering violation.
                self.conn.close()

    def read(self, size=None):
        bytes = self.resp.read(size)
        if size is None or len(bytes) < size:
            self.close()
        return bytes

    def _release_conn(self):
        self.conn_pool.release(self.url, self.conn)
        self.conn_pool, self.url, self.conn = None, None, None

    def close(self):
        while not self.resp.isclosed():
            chunk = self.resp.read(CHUNK_SIZE)
            if not chunk:
                self.resp.close()
        if self.conn:
            self._release_conn()

    def iterchunks(self):
        assert self.chunked
        buffer = []
        while True:

            if self.resp.isclosed():
                break

            chunksz = int(self.resp.fp.readline().strip(), 16)
            if not chunksz:
                self.resp.fp.read(2) #crlf
                self.resp.close()
                self._release_conn()
                break

            chunk = self.resp.fp.read(chunksz)
            for ln in chunk.splitlines(True):

                end = ln == b'\n' and not buffer # end of response
                if not ln or end:
                    break

                buffer.append(ln)
                if ln.endswith(b'\n'):
                    yield b''.join(buffer)
                    buffer = []

            self.resp.fp.read(2) #crlf


RETRYABLE_ERRORS = frozenset([
    errno.EPIPE, errno.ETIMEDOUT,
    errno.ECONNRESET, errno.ECONNREFUSED, errno.ECONNABORTED,
    errno.EHOSTDOWN, errno.EHOSTUNREACH,
    errno.ENETRESET, errno.ENETUNREACH, errno.ENETDOWN
])


class Session(object):

    def __init__(self, cache=None, timeout=None, max_redirects=5,
                 retry_delays=[0], retryable_errors=RETRYABLE_ERRORS):
        """Initialize an HTTP client session.

        :param cache: an instance with a dict-like interface or None to allow
                      Session to create a dict for caching.
        :param timeout: socket timeout in number of seconds, or `None` for no
                        timeout (the default)
        :param retry_delays: list of request retry delays.
        """
        from GRID_LRT.couchdb import __version__ as VERSION
        self.user_agent = 'CouchDB-Python/%s' % VERSION
        # XXX We accept a `cache` dict arg, but the ref gets overwritten later
        # during cache cleanup. Do we remove the cache arg (does using a shared
        # Session instance cover the same use cases?) or fix the cache cleanup?
        # For now, let's just assign the dict to the Cache instance to retain
        # current behaviour.
        if cache is not None:
            cache_by_url = cache
            cache = Cache()
            cache.by_url = cache_by_url
        else:
            cache = Cache()
        self.cache = cache
        self.max_redirects = max_redirects
        self.perm_redirects = {}

        self._disable_ssl_verification = False
        self._timeout = timeout
        self.connection_pool = ConnectionPool(
            self._timeout,
            disable_ssl_verification=self._disable_ssl_verification)

        self.retry_delays = list(retry_delays) # We don't want this changing on us.
        self.retryable_errors = set(retryable_errors)

    def disable_ssl_verification(self):
        """Disable verification of SSL certificates and re-initialize the
        ConnectionPool. Only applicable on Python 2.7.9+ as previous versions
        of Python don't verify SSL certs."""
        self._disable_ssl_verification = True
        self.connection_pool = ConnectionPool(self._timeout,
            disable_ssl_verification=self._disable_ssl_verification)

    def request(self, method, url, body=None, headers=None, credentials=None,
                num_redirects=0):
        if url in self.perm_redirects:
            url = self.perm_redirects[url]
        method = method.upper()

        if headers is None:
            headers = {}
        headers.setdefault('Accept', 'application/json')
        headers['User-Agent'] = self.user_agent

        cached_resp = None
        if method in ('GET', 'HEAD'):
            cached_resp = self.cache.get(url)
            if cached_resp is not None:
                etag = cached_resp[1].get('etag')
                if etag:
                    headers['If-None-Match'] = etag

        if (body is not None and not isinstance(body, util.strbase) and
                not hasattr(body, 'read')):
            body = json.encode(body).encode('utf-8')
            headers.setdefault('Content-Type', 'application/json')

        if body is None:
            headers.setdefault('Content-Length', '0')
        elif isinstance(body, util.strbase):
            headers.setdefault('Content-Length', str(len(body)))
        else:
            headers['Transfer-Encoding'] = 'chunked'

        authorization = basic_auth(credentials)
        if authorization:
            headers['Authorization'] = authorization

        path_query = util.urlunsplit(('', '') + util.urlsplit(url)[2:4] + ('',))
        conn = self.connection_pool.get(url)

        def _try_request_with_retries(retries):
            while True:
                try:
                    return _try_request()
                except socket.error as e:
                    ecode = e.args[0]
                    if ecode not in self.retryable_errors:
                        raise
                    try:
                        delay = next(retries)
                    except StopIteration:
                        # No more retries, raise last socket error.
                        raise e
                    finally:
                        time.sleep(delay)
                        conn.close()

        def _try_request():
            try:
                conn.putrequest(method, path_query, skip_accept_encoding=True)
                for header in headers:
                    conn.putheader(header, headers[header])
                if body is None:
                    conn.endheaders()
                else:
                    if isinstance(body, util.strbase):
                        if isinstance(body, util.utype):
                            conn.endheaders(body.encode('utf-8'))
                        else:
                            conn.endheaders(body)
                    else: # assume a file-like object and send in chunks
                        conn.endheaders()
                        while 1:
                            chunk = body.read(CHUNK_SIZE)
                            if not chunk:
                                break
                            if isinstance(chunk, util.utype):
                                chunk = chunk.encode('utf-8')
                            status = ('%x\r\n' % len(chunk)).encode('utf-8')
                            conn.send(status + chunk + b'\r\n')
                        conn.send(b'0\r\n\r\n')
                return conn.getresponse()
            except BadStatusLine as e:
                # httplib raises a BadStatusLine when it cannot read the status
                # line saying, "Presumably, the server closed the connection
                # before sending a valid response."
                # Raise as ECONNRESET to simplify retry logic.
                if e.line == '' or e.line == "''":
                    raise socket.error(errno.ECONNRESET)
                else:
                    raise

        resp = _try_request_with_retries(iter(self.retry_delays))
        status = resp.status

        # Handle conditional response
        if status == 304 and method in ('GET', 'HEAD'):
            resp.read()
            self.connection_pool.release(url, conn)
            status, msg, data = cached_resp
            if data is not None:
                data = util.StringIO(data)
            return status, msg, data
        elif cached_resp:
            self.cache.remove(url)

        # Handle redirects
        if status == 303 or \
                method in ('GET', 'HEAD') and status in (301, 302, 307):
            resp.read()
            self.connection_pool.release(url, conn)
            if num_redirects > self.max_redirects:
                raise RedirectLimit('Redirection limit exceeded')
            location = resp.getheader('location')

            # in case of relative location: add scheme and host to the location
            location_split = util.urlsplit(location)

            if not location_split[0]:
                orig_url_split = util.urlsplit(url)
                location = util.urlunsplit(orig_url_split[:2] + location_split[2:])

            if status == 301:
                self.perm_redirects[url] = location
            elif status == 303:
                method = 'GET'
            return self.request(method, location, body, headers,
                                num_redirects=num_redirects + 1)

        data = None
        streamed = False

        # Read the full response for empty responses so that the connection is
        # in good state for the next request
        if method == 'HEAD' or resp.getheader('content-length') == '0' or \
                status < 200 or status in (204, 304):
            resp.read()
            self.connection_pool.release(url, conn)

        # Buffer small non-JSON response bodies
        elif int(resp.getheader('content-length', sys.maxsize)) < CHUNK_SIZE:
            data = resp.read()
            self.connection_pool.release(url, conn)

        # For large or chunked response bodies, do not buffer the full body,
        # and instead return a minimal file-like object
        else:
            data = ResponseBody(resp, self.connection_pool, url, conn)
            streamed = True

        # Handle errors
        if status >= 400:
            ctype = resp.getheader('content-type')
            if data is not None and 'application/json' in ctype:
                data = json.decode(data.decode('utf-8'))
                error = data.get('error'), data.get('reason')
            elif method != 'HEAD':
                error = resp.read()
                self.connection_pool.release(url, conn)
            else:
                error = ''
            if status == 401:
                raise Unauthorized(error)
            elif status == 403:
                raise Forbidden(error)
            elif status == 404:
                raise ResourceNotFound(error)
            elif status == 409:
                raise ResourceConflict(error)
            elif status == 412:
                raise PreconditionFailed(error)
            else:
                raise ServerError((status, error))

        # Store cachable responses
        if not streamed and method == 'GET' and 'etag' in resp.msg:
            self.cache.put(url, (status, resp.msg, data))

        if not streamed and data is not None:
            data = util.StringIO(data)

        return status, resp.msg, data


def cache_sort(i):
    return datetime.fromtimestamp(time.mktime(parsedate(i[1][1]['Date'])))

class Cache(object):
    """Content cache."""

    # Some random values to limit memory use
    keep_size, max_size = 10, 75

    def __init__(self):
        self.by_url = {}

    def get(self, url):
        return self.by_url.get(url)

    def put(self, url, response):
        self.by_url[url] = response
        if len(self.by_url) > self.max_size:
            self._clean()

    def remove(self, url):
        self.by_url.pop(url, None)

    def _clean(self):
        ls = sorted(self.by_url.items(), key=cache_sort)
        self.by_url = dict(ls[-self.keep_size:])


class InsecureHTTPSConnection(HTTPSConnection):
    """Wrapper class to create an HTTPSConnection without SSL verification
    (the default behavior in Python < 2.7.9). See:
    https://docs.python.org/2/library/httplib.html#httplib.HTTPSConnection"""
    if sys.version_info >= (2, 7, 9):
        def __init__(self, *a, **k):
            k['context'] = ssl._create_unverified_context()
            HTTPSConnection.__init__(self, *a, **k)


class ConnectionPool(object):
    """HTTP connection pool."""

    def __init__(self, timeout, disable_ssl_verification=False):
        self.timeout = timeout
        self.disable_ssl_verification = disable_ssl_verification
        self.conns = {} # HTTP connections keyed by (scheme, host)
        self.lock = Lock()

    def get(self, url):

        scheme, host = util.urlsplit(url, 'http', False)[:2]

        # Try to reuse an existing connection.
        self.lock.acquire()
        try:
            conns = self.conns.setdefault((scheme, host), [])
            if conns:
                conn = conns.pop(-1)
            else:
                conn = None
        finally:
            self.lock.release()

        # Create a new connection if nothing was available.
        if conn is None:
            if scheme == 'http':
                cls = HTTPConnection
            elif scheme == 'https':
                if self.disable_ssl_verification:
                    cls = InsecureHTTPSConnection
                else:
                    cls = HTTPSConnection
            else:
                raise ValueError('%s is not a supported scheme' % scheme)
            conn = cls(host, timeout=self.timeout)
            conn.connect()

        return conn

    def release(self, url, conn):
        scheme, host = util.urlsplit(url, 'http', False)[:2]
        self.lock.acquire()
        try:
            self.conns.setdefault((scheme, host), []).append(conn)
        finally:
            self.lock.release()

    def __del__(self):
        for key, conns in list(self.conns.items()):
            for conn in conns:
                conn.close()


class Resource(object):

    def __init__(self, url, session, headers=None):
        if sys.version_info[0] == 2 and isinstance(url, util.utype):
            url = url.encode('utf-8') # kind of an ugly hack for issue 235
        self.url, self.credentials = extract_credentials(url)
        if session is None:
            session = Session()
        self.session = session
        self.headers = headers or {}

    def __call__(self, *path):
        obj = type(self)(urljoin(self.url, *path), self.session)
        obj.credentials = self.credentials
        obj.headers = self.headers.copy()
        return obj

    def delete(self, path=None, headers=None, **params):
        return self._request('DELETE', path, headers=headers, **params)

    def get(self, path=None, headers=None, **params):
        return self._request('GET', path, headers=headers, **params)

    def head(self, path=None, headers=None, **params):
        return self._request('HEAD', path, headers=headers, **params)

    def post(self, path=None, body=None, headers=None, **params):
        return self._request('POST', path, body=body, headers=headers,
                             **params)

    def put(self, path=None, body=None, headers=None, **params):
        return self._request('PUT', path, body=body, headers=headers, **params)

    def delete_json(self, path=None, headers=None, **params):
        return self._request_json('DELETE', path, headers=headers, **params)

    def get_json(self, path=None, headers=None, **params):
        return self._request_json('GET', path, headers=headers, **params)

    def post_json(self, path=None, body=None, headers=None, **params):
        return self._request_json('POST', path, body=body, headers=headers,
                                  **params)

    def put_json(self, path=None, body=None, headers=None, **params):
        return self._request_json('PUT', path, body=body, headers=headers,
                                  **params)

    def _request(self, method, path=None, body=None, headers=None, **params):
        all_headers = self.headers.copy()
        all_headers.update(headers or {})
        if path is not None:
            if isinstance(path, list):
                url = urljoin(self.url, *path, **params)
            else:
                url = urljoin(self.url, path, **params)
        else:
            url = urljoin(self.url, **params)
        return self.session.request(method, url, body=body,
                                    headers=all_headers,
                                    credentials=self.credentials)

    def _request_json(self, method, path=None, body=None, headers=None, **params):
        status, headers, data = self._request(method, path, body=body,
                                              headers=headers, **params)
        if 'application/json' in headers.get('content-type', ''):
            data = json.decode(data.read().decode('utf-8'))
        return status, headers, data



def extract_credentials(url):
    """Extract authentication (user name and password) credentials from the
    given URL.

    >>> extract_credentials('http://localhost:5984/_config/')
    ('http://localhost:5984/_config/', None)
    >>> extract_credentials('http://joe:secret@localhost:5984/_config/')
    ('http://localhost:5984/_config/', ('joe', 'secret'))
    >>> extract_credentials('http://joe%40example.com:secret@localhost:5984/_config/')
    ('http://localhost:5984/_config/', ('joe@example.com', 'secret'))
    """
    parts = util.urlsplit(url)
    netloc = parts[1]
    if '@' in netloc:
        creds, netloc = netloc.split('@')
        credentials = tuple(util.urlunquote(i) for i in creds.split(':'))
        parts = list(parts)
        parts[1] = netloc
    else:
        credentials = None
    return util.urlunsplit(parts), credentials


def basic_auth(credentials):
    """Generates authorization header value for given credentials.
    >>> basic_auth(('root', 'relax'))
    b'Basic cm9vdDpyZWxheA=='
    >>> basic_auth(None)
    >>> basic_auth(())
    """
    if credentials:
        token = b64encode(('%s:%s' % credentials).encode('utf-8'))
        return ('Basic %s' % token.strip().decode('utf-8')).encode('ascii')


def quote(string, safe=''):
    if isinstance(string, util.utype):
        string = string.encode('utf-8')
    return util.urlquote(string, safe)


def urlencode(data):
    if isinstance(data, dict):
        data = data.items()
    params = []
    for name, value in data:
        if isinstance(value, util.utype):
            value = value.encode('utf-8')
        params.append((name, value))
    return util.urlencode(params)


def urljoin(base, *path, **query):
    """Assemble a uri based on a base, any number of path segments, and query
    string parameters.

    >>> urljoin('http://example.org', '_all_dbs')
    'http://example.org/_all_dbs'

    A trailing slash on the uri base is handled gracefully:

    >>> urljoin('http://example.org/', '_all_dbs')
    'http://example.org/_all_dbs'

    And multiple positional arguments become path parts:

    >>> urljoin('http://example.org/', 'foo', 'bar')
    'http://example.org/foo/bar'

    All slashes within a path part are escaped:

    >>> urljoin('http://example.org/', 'foo/bar')
    'http://example.org/foo%2Fbar'
    >>> urljoin('http://example.org/', 'foo', '/bar/')
    'http://example.org/foo/%2Fbar%2F'

    >>> urljoin('http://example.org/', None) #doctest:+IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    TypeError: argument 2 to map() must support iteration
    """
    if base and base.endswith('/'):
        base = base[:-1]
    retval = [base]

    # build the path
    path = '/'.join([''] + [quote(s) for s in path])
    if path:
        retval.append(path)

    # build the query string
    params = []
    for name, value in query.items():
        if type(value) in (list, tuple):
            params.extend([(name, i) for i in value if i is not None])
        elif value is not None:
            if value is True:
                value = 'true'
            elif value is False:
                value = 'false'
            params.append((name, value))
    if params:
        retval.extend(['?', urlencode(params)])

    return ''.join(retval)





__all__ = [
    'StringIO', 'urlsplit', 'urlunsplit', 'urlquote', 'urlunquote',
    'urlencode', 'utype', 'btype', 'ltype', 'strbase', 'funcode', 'urlparse',
]

utype = unicode
btype = str
ltype = long
strbase = str, bytes, unicode

from io import BytesIO as StringIO
from urlparse import urlparse, urlsplit, urlunsplit
from urllib import quote as urlquote
from urllib import unquote as urlunquote
from urllib import urlencode


def funcode(fun):
    return fun.func_code




__all__ = [
    'StringIO', 'urlsplit', 'urlunsplit', 'urlquote', 'urlunquote',
    'urlencode', 'utype', 'btype', 'ltype', 'strbase', 'funcode', 'urlparse',
]

utype = str
btype = bytes
ltype = int
strbase = str, bytes

from io import BytesIO as StringIO
from urllib.parse import urlsplit, urlunsplit, urlencode, urlparse
from urllib.parse import quote as urlquote
from urllib.parse import unquote as urlunquote


def funcode(fun):
    return fun.__code__



#!/usr/bin/env python
"""Setup script for the Grid LOFAR Tools (GRID_LRT) python module
"""
#from distutils.core import setup
from setuptools import setup
import GRID_LRT

setup(name='GRID_LRT',
      packages=['GRID_LRT', 'GRID_LRT/Staging', 'GRID_LRT/application','GRID_LRT/storage', 'GRID_LRT/auth'],
      version=GRID_LRT.__version__,
      setup_requires=[
          'pyyaml',
          ],
      tests_require=[
          'pytest',
          ],
      install_requires=[
          'pyyaml',
          'couchdb',
          'retrying',
          'humanfriendly'
          ],
      include_package_data=True,
      data_files=[("githash",["GRID_LRT/__githash__"]),
                 ("GRID_LRT/data/config",
                   ['GRID_LRT/data/config/bash_file.cfg',
                    'GRID_LRT/data/config/NDPPP_parset.cfg',
                    "GRID_LRT/data/config/tutorial.cfg",
                    "GRID_LRT/data/config/steps/pref_cal1.cfg",
                    "GRID_LRT/data/config/steps/pref_cal2.cfg",
                    "GRID_LRT/data/config/steps/pref_targ1.cfg",
                    "GRID_LRT/data/config/steps/pref_targ2.cfg"]),
                  ("GRID_LRT/data/launchers/",
                   ["GRID_LRT/data/launchers/run_remote_sandbox.sh"])
                 ],
      description='GRID LOFAR Reduction Tools',
      long_description="Software that encapsulates LOFAR processing"
      "and allows the use of High Throughput Distributed processing available"
      "at SURFsara and other European Grid Initiative locations.",
      author='Alexandar Mechev',
      author_email='LOFAR@apmechev.com',
      url='https://www.github.com/apmechev/GRID_LRT/',
      download_url='https://github.com/apmechev/GRID_LRT/archive/v0.2.0.tar.gz',
      keywords=['surfsara', 'distributed-computing', 'LOFAR'],
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                   "Natural Language :: English",
                   "Topic :: Scientific/Engineering :: Astronomy",
                   "Topic :: System :: Distributed Computing",
                   'Programming Language :: Python :: 2.6',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.2',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4']
     )




# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Alexandar P. Mechev
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.md, which
# you should have received as part of this distribution.

"""
.. module:: GRID_LRT.Token
   :platform: Unix
   :synopsis: Set of tools for manually and automatically creating tokens

.. moduleauthor:: Alexandar Mechev <LOFAR@apmechev.com>
.. note:: Will be renamed GRID_LRT.token to conform to python standards

>>> #Example creation of a token of token_type 'test'
>>> from GRID_LRT.auth.get_picas_credentials import picas_cred
>>> pc=picas_cred() #Gets picas_credentials
>>>
>>> th=Token.Token_Handler( t_type="test", uname=pc.user, pwd=pc.password, dbn=pc.database)
>>> #creates object to 'handle' Tokens
>>> th.add_overview_view()
>>> th.add_status_views() #Adds 'todo', 'done', 'locked' and 'error' views
>>> th.load_views()
>>> th.views.keys()
>>> th.reset_tokens(view_name='error') # resets all tokens in 'error' view
>>> th.set_view_to_status(view_name='done','processed')
"""

from __future__ import print_function
import sys
import os
import shutil
import itertools
import time
import tarfile
import yaml
from retrying import retry


import GRID_LRT
from GRID_LRT.couchdb.design import ViewDefinition
if 'couchdb' not in sys.modules:
    from GRID_LRT import couchdb

__version__ = GRID_LRT.__version__
__author__ = GRID_LRT.__author__
__license__ = GRID_LRT.__license__
__email__ = GRID_LRT.__email__
__copyright__ = GRID_LRT.__copyright__
__credits__ = GRID_LRT.__credits__
__maintainer__ = GRID_LRT.__maintainer__
__status__ = GRID_LRT.__status__

def get_all_design_docs(pcreds = None, srv="https://picas-lofar.grid.surfsara.nl:6984"):
    """Returns a list of design documents for the pcreds.databse on server=srv
    If pcreds are none, then we're adminparty and db is test_db"""
    if pcreds:
        user, passwd, dbn = pcreds.user, pcreds.password, pcreds.database
    else:
        user, passwd, dbn = "","","test_db"
    server = couchdb.Server(srv)
    if user and passwd:
        server.resource.credentials = (user, passwd)
    database = server[dbn]
    ad=[doc for doc in database.get('_all_docs')['rows'] if '_design' in doc['id']]
    return [i['id'] for i in ad]


def reset_all_tokens(token_type, picas_creds, server="https://picas-lofar.grid.surfsara.nl:6984"):
    """ Resets all Tokens with the pc authorization
    """
    thandler = Token_Handler(t_type=token_type, srv=server,
                             uname=picas_creds.user, pwd=picas_creds.password,
                             dbn=picas_creds.database)
    thandler.load_views()
    for view in list(thandler.views):
        if view != 'overview_total':
            thandler.reset_tokens(view)


def purge_tokens(token_type, picas_creds, server="https://picas-lofar.grid.surfsara.nl:6984"):
    """Automated function to purge tokens authorizing with Picas_creds"""
    thandler = Token_Handler(t_type=token_type, srv=server,
                             uname=picas_creds.user, pwd=picas_creds.password,
                             dbn=picas_creds.database)
    thandler.load_views()
    thandler.purge_tokens()


class Token_Handler(object):
    """

    The Token_Handler class uses couchdb to create, modify and delete
    tokens and views, to attach files, or download attachments and to
    easily modify fields in tokens. It's initiated with the token_type,
    server, username, password and name of database.

    """

    def __init__(self, t_type="token",
                 srv="https://picas-lofar.grid.surfsara.nl:6984",
                 uname="", pwd="", dbn=""):
        """
        >>> #Example creation of a token of token_type 'test'
        >>> from GRID_LRT.auth.get_picas_credentials import picas_cred
        >>> pc=picas_cred() #Gets picas_credentials
        >>>
        >>> th=Token.Token_Handler( t_type="test",
        srv="https://picas-lofar.grid.surfsara.nl:6984", uname=pc.user,
        pwd=pc.password, dbn=pc.database ) #creates object to 'handle' Tokens
        >>> th.add_overview_view()
        >>> th.add_status_views() #Adds 'todo', 'done', 'locked' and 'error' views
        >>> th.load_views()
        >>> th.views.keys()
        >>> th.reset_tokens(view_name='error') # resets all tokens in 'error' view
        >>> th.set_view_to_status(view_name='done','processed')
        """
        if t_type:
            self.t_type = t_type
        else:
            raise Exception("t_type not defined!")
        self.database = self._get_db(uname, pwd, dbn, srv)
        self.views = {}
        self.tokens = {}

    @staticmethod
    def _get_db(uname, pwd, dbn, srv):
        """Logs into the Couchdb server and returns the database requested.
        Returns a couchDB database object

        :param uname: The username to log into CouchDB with
        :type uname: str
        :param pwd: The CouchDB password
        :type pwd: str
        :param dbn: The CouchDB Database Name
        :type dbn: str
        :param srv: URL of the CouchDB Server
        :type srv: str
        :return: A CouchDB Database Object
        :rtype: :class:`~couchdb.client.Database`
        """
        server = couchdb.Server(srv)
        if uname and pwd:
            server.resource.credentials = (uname, pwd)
        database = server[dbn]
        return database

    @retry(wait_fixed=20, stop_max_attempt_number=2)
    def create_token(self, keys=None, append="", attach=None):
        '''Creates a token, appends string to token ID if requested and
        adds user requested keys through the dict keys{}

        :param keys: A dictionary of keys, which will be uploaded to the CouchDB document.
            The supported values for a key are str,int,float and dict
        :type keys: dict
        :param append: A string which is appended to the end of the tokenID, useful for
            adding an OBSID for example
        :type append: str
        :param attach: A 2-item list of file to be attached to the token.
        The first value is the file handle and the second is a string with
        the attachment name. ex: [open('/home/apmechev/file.txt','r'),"file.txt"]
        :type attach: list
        :return: A string with the token ID
        :rtype: str
        '''
        default_keys = {
            '_id': 't_'+self.t_type+"_",
            'type': self.t_type,
            'lock': 0,
            'done': 0,
            'hostname': '',
            'scrub_count': 0,
            'output': "",
            'created': time.time()
        }
        if keys:
            keys = dict(itertools.chain(keys.items(), default_keys.items()))
            self._append_id(keys, append)
        else:
            self._append_id(default_keys, append)
        self.tokens[keys["_id"]] = keys
        self.database.update([keys])
        if attach:
            self.add_attachment(keys['_id'], attach[0], attach[1])
        return keys['_id']  # returns the token ID

    @staticmethod
    def _append_id(keys, app=""):
        """ Helper function that appends a string to the token ID"""
        keys["_id"] += app

    def load_views(self):
        """Helper function to get the current views on the database.
        Updates the internal self.views variable
        """
        db_views = self.database.get("_design/"+self.t_type)
        if db_views is None:
            RuntimeWarning("No views found in design document")
            return
        self.views = db_views["views"]

    def delete_tokens(self, view_name="test_view", key=None):
        """Deletes tokens from view view_name

            exits if the view doesn't exist

            User can select which tokens within the view to delete

            >>> t1.delete_tokens("todo",["OBSID","L123456"])
            >>> #Only deletes tokens with OBSID key = L123456
            >>> t1.delete_tokens("error") # Deletes all error tokens

        :param view_name: Name of the view from which to delete tokens
        :type view_name: str
        :param key: key-value pair that selects which tokens to delete
        (by default empty == delete all token)
        :type key: list
        """
        view = self.list_tokens_from_view(view_name)
        to_delete = []
        for tok in view:
            document = self.database[tok['key']]
            if not key:
                pass
            else:
                if document[key[0]] != key[1]:
                    continue
            print("Deleting Token "+tok['id'])
            to_delete.append(document)
        self.database.purge(to_delete)

    def add_view(self, view_name="test_view",
                 cond='doc.lock > 0 && doc.done > 0 && doc.output < 0 ',
                 emit_value='doc._id', emit_value2='doc._id'):
        """Adds a view to the db, needs a view name and a condition.
        Emits all tokens with the type of Token_Handler.t_type, that also match the condition

        :param view_name: The name of the new view to be created
        :type view_name: str
        :param cond: A string containing the condition which all tokens of the view must match.
        It can include boolean operators '>', '<'and '&&'.
        The token's fields are refered to as 'doc.field'
        :type cond: str
        :param emit_value: The (first) emit value that is returne by the view.
        If you look on couchdb/request the tokens from the view, you'll get two values.
        This will be the first (typically the token's ID)
        :type emit_value: str
        :param emit_value2: The second emit value. You can thus return the token's ID
        and its status for example
        :type emit_value2: str
        """
        general_view_code = '''
        function(doc) {
           if(doc.type == "%s") {
            if(%s) {
              emit(%s, %s );
            }
          }
        }
        '''
        view = ViewDefinition(self.t_type, view_name, general_view_code % (
            self.t_type, cond, emit_value, emit_value2))
        self.views[view_name] = view
        view.sync(self.database)

    def add_mapreduce_view(self, view_name="test_mapred_view",
                           cond='doc.PIPELINE_STEP == "pref_cal1" '):
        """
        While the overview_view is applied to all the tokens in the design document,
        this 'mapreduce' view is useful if instead of regular view, you want to filter
        the tokens and display the user with a 'mini-oververview' view.
        This way you can check the status of a subset of the tokens.

        """
        overview_map_code = '''
function(doc) {
   if(doc.type == "%s" )
      if(%s){
        {
       if (doc.lock == 0 && doc.done == 0){
          emit('todo', 1);
       }
       if(doc.lock > 0 && doc.status == 'downloading' ) {
          emit('downloading', 1);
       }
       if(doc.lock > 0 && doc.done > 0 && doc.output == 0 ) {
          emit('done', 1);
       }
       if(doc.lock > 0 && doc.output != 0 && doc.output != "" ) {
          emit('error', 1);
       }
       if(doc.lock > 0 && doc.status == 'launched' ) {
          emit('waiting', 1);
       }
       if(doc.lock > 0  && doc.done==0 && doc.status!='downloading' ) {
          emit('running', 1);
       }
     }
   }
}
'''
        overview_reduce_code = '''
function (key, values, rereduce) {
   return sum(values);
}
'''
        overview_total_view = ViewDefinition(self.t_type, view_name,
                                             overview_map_code % (
                                                 self.t_type, cond),
                                             overview_reduce_code)
        self.views['overview_total'] = overview_total_view
        overview_total_view.sync(self.database)

    def add_overview_view(self):
        """ Helper function that creates the Map-reduce view which makes it easy to count
            the number of jobs in the 'locked','todo','downloading','error' and 'running' states
        """
        overview_map_code = '''
function(doc) {
   if(doc.type == "%s") {
       if (doc.lock == 0 && doc.done == 0){
          emit('todo', 1);
       }
       if(doc.lock > 0 && doc.status == 'downloading' ) {
          emit('downloading', 1);
       }
       if(doc.lock > 0 && doc.done > 0 && doc.output == 0 ) {
          emit('done', 1);
       }
       if(doc.lock > 0 && doc.output != 0 ) {
          emit('error', 1);
       }
       if(doc.lock > 0 && doc.status == 'launched' ) {
          emit('waiting', 1);
       }
       if(doc.lock > 0  && doc.done==0 && doc.status!='downloading' ) {
          emit('running', 1);
       }
   }
}
'''
        overview_reduce_code = '''
function (key, values, rereduce) {
   return sum(values);
}
'''
        overview_total_view = ViewDefinition(self.t_type, 'overview_total',
                                             overview_map_code % (self.t_type),
                                             overview_reduce_code)
        self.views['overview_total'] = overview_total_view
        overview_total_view.sync(self.database)

    def add_status_views(self):
        """ Adds the 'todo', locked, done and error views. the TODO view is necessary for the
        worker node to find an un-locked token
        """
        self.add_view(view_name="todo",
                      cond='doc.lock ==  0 && doc.done == 0 ')
        self.add_view(view_name="locked",
                      cond='doc.lock > 0 && doc.done == 0 ')
        self.add_view(view_name="done", cond='doc.status == "done" ')
        self.add_view(view_name="error", cond='doc.output != 0 ',
                      emit_value2='doc.output')

    def del_view(self, view_name="test_view"):
        '''Deletes the view with view name from the _design/${token_type} document
            and from the token_Handler's dict of views

        :param view_name: The name of the view which should be removed
        :type view_name: str
        '''
        db_views = self.database.get("_design/"+self.t_type)
        db_views["views"].pop(view_name, None)
        self.views.pop(view_name, None)
        self.database.update([db_views])

    def remove_error(self):
        ''' Removes all tokens in the error view
        '''
        cond = "doc.lock > 0 && doc.done > 0 && doc.output > 0"
        self.add_view(view_name="error", cond=cond)
        self.delete_tokens("error")

    def reset_tokens(self, view_name="test_view", key=None, del_attach=False):
        """ resets all tokens in a view, optionally can reset all tokens in a view
            who have key-value pairs matched by key[0],key[1]

            >>> t1.reset_token("error")
            >>> t1.reset_token("error",key=["OBSID","L123456"])
            >>> t1.reset_token("error",key=["scrub_count",6])
        """
        view = self.list_tokens_from_view(view_name)
        to_update = []
        for tok in view:
            document = self.database[tok['key']]

            if key and document[key[0]] != key[1]:  # make it not just equal
                continue
            try:
                document['status'] = 'todo'
            except KeyError:
                pass
            document['lock'] = 0
            document['done'] = 0
            document['scrub_count'] += 1
            document['hostname'] = ''
            document['output'] = ''
            if del_attach:
                if "_attachments" in document:
                    del document["_attachments"]
            to_update.append(document)
        self.database.update(to_update)
        return to_update

    def add_attachment(self, token, filehandle, filename="test"):
        """Uploads an attachment to a token

        Args:
            :param token: The Token _id as recorded in the CouchDB database
            :type token: str
            :param filehandle: the file handle to the file which to upload open(filepath,'r')
            :type tok_config: os.file()
            :param filename: The name of the attachment
            :type tok_config: str

        """
        self.database.put_attachment(self.database[token], filehandle, filename)

    def list_attachments(self, token):
        """Lists all of the filenames attached to a couchDB token
        """
        tok_keys = self.database[token]
        if '_attachments' in tok_keys:
            return list(tok_keys["_attachments"].keys())
        return []

    def get_attachment(self, token, filename, savename=None):
        """Downloads an attachment from a CouchDB token. Optionally
            a save name can be specified.
        """
        try:
            attach = self.database.get_attachment(token, filename).read()
        except AttributeError:
            print("error getting attachment: "+str(filename))
            return ""
        if savename is not None:
            savefile = savename
        else:
            savefile = filename
        if "/" in filename:
             savefile = filename.replace("/", "_")
        with open(savefile, 'w') as _file:
            for line in attach:
                _file.write(str(line))
        return os.path.abspath(savefile)

    def list_tokens_from_view(self, view_name):
        """Returns all tokens from a viewname
        """
        self.load_views()
        if view_name in self.views:
            view = self.views[view_name]
        else:
            RuntimeWarning("View Named "+view_name+" Doesn't exist")
            return None
        view = self.database.view(self.t_type+"/"+view_name)
        return view

    def archive_tokens_from_view(self, viewname, delete_on_save=False):
        to_del = []
        for token in self.list_tokens_from_view(viewname):
            self.archive_a_token(token['id'], delete_on_save)
            to_del.append(self.database[token['id']])
        if delete_on_save:
            self.database.purge(to_del)

    def archive_tokens(self, delete_on_save=False, compress=True):
        """Archives all tokens and attachments into a folder

        """
        os.mkdir(self.t_type)
        os.chdir(self.t_type)
        self.load_views()
        for view in self.views.keys():
            if view == 'overview_total':
                continue
            self.archive_tokens_from_view(view, delete_on_save)
        result_link = os.getcwd()
        os.chdir('..')
        if compress:
            with tarfile.open(name="tokens_"+self.t_type+".tar.gz", mode='w:gz') as t_file:
                t_file.add(self.t_type+"/")
                result_link = t_file.name
                shutil.rmtree(self.t_type+"/")
        if delete_on_save:
            self.purge_tokens()
        return result_link

    def archive_a_token(self, tokenid, delete=False):
        """Dumps the token data into a yaml file and saves the attachments
        returns list of archived files"""
        data = self.database[tokenid]
        archived_files = []
        with open(tokenid+".dump", 'w') as savefile:
            yaml.dump(data, savefile)
        archived_files.append(tokenid+".dump")
        for att_file in self.list_attachments(tokenid):
            fname = att_file.replace('/', '-')
            self.get_attachment(tokenid, att_file, tokenid +
                                "_attachment_"+str(fname))
            archived_files.append(fname)
        if delete: #test
            self.database.purge([data])
        return archived_files 

    def clear_all_views(self):
        """Iterates over all views in the design document
        and deletes all tokens from those views. Finally, removes
        the views from the database"""
        self.load_views()
        for view in list(self.views):
            if view != 'overview_total':
                self.delete_tokens(view)
            self.del_view(view)
        self.load_views()
        return self.views

    def purge_tokens(self):
        """ Deletes ALL tokens associated with this token_type
        and removes all views. Also removes the design document from
        the database"""
        self.clear_all_views()
        del self.database['_design/'+self.t_type]


    def set_view_to_status(self, view_name, status):
        """Sets the status to all tokens in 'view' to 'status
            eg. Set all locked tokens to error or all error tokens to todo
            it also locks the tokens!
        """
        view = self.list_tokens_from_view(view_name)
        to_update = []
        for token in view:
            document = self.database[token['key']]
            document['status'] = str(status)
            document['lock'] = 1
            to_update.append(document)
        self.database.update(to_update)


class TokenSet(object):
    """ The TokenSet object can automatically create a group of tokens from a
    yaml configuration file and a dictionary. It keeps track internally of
    the set of tokens and allows users to batch attach files to the entire TokenSet or alter
    fields of all tokens in the set.

    """

    def __init__(self, th=None, tok_config=None):
        """ The TokenSet object is created with a TokenHandler Object, which is
        responsible for the interface to the CouchDB views and Documents. This also ensures
        that only one job type is contained in a TokenSet.

        Args:
            :param th: The TokenHandler associated with the job tokens
            :type th: GRID_LRT.Token.TokenHandler
            :param tok_config: Location of the token yaml file on the host FileSystem
            :type tok_config: str
            :raises: AttributeError, KeyError

        """
        self.thandler = th
        self.__tokens = []
        if not tok_config:
            self.token_keys = {}
        else:
            with open(tok_config, 'r') as optfile:
                self.token_keys = yaml.load(optfile)['Token']

    def create_dict_tokens(self, iterable={}, id_prefix='SB', id_append="L000000",
                           key_name='STARTSB', file_upload=None):
        """ A function that accepts a dictionary and creates a set of tokens equal to
            the number of entries (keys) of the dictionary. The values of the dict are
            a list of strings that may be attached to each token if the 'file_upload'
            argument exists.

            Args:
                :param iterable: The dictionary which determines how many tokens will be created.
                The values  are attached to each token
                :type iterable: dict
                :param id_append: Option to append the OBSID to each Token
                :type id_append: str
                :param key_name: The Token field which will hold the value of the dictionary's
                keys for each Token
                :type key_name: str
                :param file_upload: The name of the file which to upload to the tokens
                (typically srm.txt)
                :type file_upload: str

        """  # TODO: Check if key_name is inside token_keys!
        for key in iterable:
            keys = dict(itertools.chain(self.token_keys.items(), {
                key_name: str("%03d" % int(key))}.items()))
#            _=keys.pop('_attachments')
            pipeline = ""
            if 'PIPELINE_STEP' in keys:
                pipeline = "_"+keys['PIPELINE_STEP']
            token = self.thandler.create_token(
                keys, append=id_append+pipeline+"_"+id_prefix+str("%03d" % int(key)))
            if file_upload:
                with open('temp_abn', 'w') as tmp_abn_file:
                    for i in iterable[key]:
                        tmp_abn_file.write("%s\n" % i)
                with open('temp_abn', 'r') as tmp_abn_file:
                    self.thandler.add_attachment(token, tmp_abn_file, file_upload)
                os.remove('temp_abn')
            self.__tokens.append(token)

    def add_attach_to_list(self, attachment, tok_list=None, name=None):
        '''Adds an attachment to all the tokens in the TokenSet, or to another list
            of tokens if explicitly specified.
        '''
        if not name:
            name = attachment
        if not tok_list:
            tok_list = self.__tokens
        for token in tok_list:
            self.thandler.add_attachment(token, open(
                attachment, 'r'), os.path.basename(name))

    @property
    def tokens(self):
        self.update_local_tokens()
        return self.__tokens

    def update_local_tokens(self):
        self.__tokens = []
        self.thandler.load_views()
        for view in self.thandler.views.keys():
            if view != 'overview_total':
                for token in self.thandler.list_tokens_from_view(view):
                    self.__tokens.append(token['id'])

    def add_keys_to_list(self, key, val, tok_list=None):
        if not tok_list:
            tok_list = self.__tokens
        to_update = []
        for token in tok_list:
            document = self.thandler.database[token]
            document[key] = str(val)
            to_update.append(document)
        self.thandler.database.update(to_update)



"""
Sandbox building and uploading module
"""

from __future__ import print_function

import os
import shutil
import tempfile
import subprocess
import warnings
warnings.simplefilter('default')
import yaml
import GRID_LRT
from GRID_LRT.auth import grid_credentials
import warnings

class Sandbox(object):
    """ A set of functions to create a sandbox from a configuration file. Uploads to grid storage
        and ssh-copies to a remote ftp server as a fallback location.

        Usage with a .cfg file:

        >>> from GRID_LRT import sandbox
        >>> s=sandbox.Sandbox()
        >>> s.build_sandbox('GRID_LRT/data/config/bash_file.cfg')
        >>> s.upload_sandbox()

        This will build the sandbox according to the recipe in bash_file.cfg and upload it to grid
        storage
    """

    def __init__(self, cfgfile=None, **kwargs):
        """ Creates a 'sandbox' object which builds and uploads the sanbox. An optional
        argument is the configuration file which is a yaml file specifying the repositories
        to include, the type of the sanbox, and its name.

        Example configuration files are included in GRID_LRT/data/config.

        :param cfgfile: The name of the configuration file to build a sandbox from
        :type cfgfile: str


        """
        self.authorized = False
        if 'authorize' in kwargs.keys() and kwargs['authorize'] == False:
            pass
        else:
            grid_credentials.grid_credentials_enabled()
        self.authorized = True
        lrt_module_dir = os.path.abspath(
            GRID_LRT.__file__).split("__init__.py")[0]
        self.base_dir = lrt_module_dir+"data/"
        self.return_dir = os.getcwd()
        self.sbxloc = None
        if cfgfile:
            self.parseconfig(cfgfile)
    
    def __enter__(self):
        pass

    def __exit__(self, ex_type, ex_value, ex_traceback):
        pass
        #        if 'remove_when_done' in self.sbx_def.keys():
#            if self.sbx_def['remove_when_done'] is True:
#                self.cleanup()

    def parseconfig(self, yamlfile):
        """Helper function to parse the sandbox configuration options
        from the yaml .cfg file. Loads the options in a dictionary
        stored in an internal variable

        :param yamlfile: The name of the sandbox configuration file
        :type yamlfile: str

        """
        with open(yamlfile, 'r') as optfile:
            opts_f = yaml.load(optfile)
        self.sbx_def = opts_f['Sandbox']
        self.shell_vars = opts_f['Shell_variables']
        self.tok_vars = opts_f['Token']

    def create_sbx_folder(self):
        '''Makes an empty sandbox folder or removes previous one
        '''
        sbx_dir = tempfile.mkdtemp()
        self.tmpdir = sbx_dir
        if not os.path.exists(self.tmpdir):
            os.makedirs(self.tmpdir)
        else:
            shutil.rmtree(self.tmpdir)
            os.makedirs(self.tmpdir)
        self.enter_sbx_folder(self.tmpdir)
        self.sbxloc = self.tmpdir
        return self.tmpdir

    def delete_sbx_folder(self):
        '''Removes the sandbox folder and subfolders
        '''
        if os.path.basename(os.getcwd()) == self.sbx_def['name']:
            os.chdir(self.base_dir)
        if os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def enter_sbx_folder(self, directory=None):
        """Changes directory to the (temporary) sandbox folder)"""
        sbx_dir = directory if directory else self.sbx_def['name']
        if os.path.exists(self.base_dir+sbx_dir):
            os.chdir(self.base_dir+sbx_dir)

    def load_git_scripts(self):
        '''Loads the git scripts into the sandbox folder. Top dir names
            are defined in the yaml, not by the git name
        '''
        if os.path.basename(os.getcwd()) != self.tmpdir:
            self.enter_sbx_folder(self.tmpdir)
        gits = self.sbx_def['git_scripts']
        if not gits:
            return
        for git in gits:
            clone = subprocess.Popen(
                ['git', 'clone', gits[git]['git_url'], self.tmpdir+"/"+git])
            clone.wait()
            os.chdir(self.tmpdir+"/"+git)
            if 'branch' in self.sbx_def['git_scripts'][git].keys():
                checkout = subprocess.Popen(
                    ['git', 'checkout', gits[git]['branch']])
                checkout.wait()
            if 'commit' in self.sbx_def['git_scripts'][git].keys():
                checkout = subprocess.Popen(
                    ['git', 'checkout', gits[git]['commit']])
                checkout.wait()
            shutil.rmtree('.git/')
            os.chdir(self.tmpdir+"/")

    def copy_git_scripts(self):
        """ Reads the location of the sandbox base scripts repository
        and clones in the current directory. Checks out the appropriate branch
        """
        print("Checking out Sanbox repository")
        subprocess.call(
            'git clone   ' + self.sbx_def['git']['location'] + " " + self.tmpdir, shell=True)
        os.chdir(self.tmpdir)
        checkout = subprocess.Popen(
            ['git', 'checkout', self.sbx_def['git']['branch']])
        checkout.wait()


    def copy_base_scripts(self, basetype=None):
        """Backwards compatible"""
        if 'git' not in self.sbx_def.keys():
            raise RuntimeError("No github repository for the sandbox inside the cfg file!")
        self.copy_git_scripts()


    def upload_sbx(self, loc=None, upload_name=None):
        """Uploads sandbox to all possible locations """
#        self.upload_gsi_sbx(loc, upload_name)
        self.upload_gsi_sbx(upload_name=upload_name,
                            loc=('gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/'
                                 'data/lofar/user/sksp/distrib/sandbox'))

    def upload_ssh_sandbox(self, upload_name=None):
        gsiloc = '/disks/ftphome/pub/apmechev/sandbox/'
        rename = self.sbxtarfile

        if not upload_name:
            if  ".tar" not in rename:
                rename = rename+".tar"
            upload_name = rename

        if self.sbxtarfile:
            upload = subprocess.Popen(['scp', self.sbxtarfile, "gaasp:"+gsiloc +
                                       self.sbx_def['loc']+"/"+upload_name])
            upload.wait()


    def upload_gsi_sbx(self, loc=None, upload_name=None):
        """ Uploads the sandbox to the relative folders
        """
        gsiloc = ("gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl"
                  "/data/lofar/user/sksp/sandbox/")

        rename = self.sbxtarfile

        if not upload_name:
            if ".tar" not in rename:
                rename = rename+".tar"
            upload_name = rename

        upload_place = gsiloc+self.sbx_def['loc']
        if loc is not None:
            upload_place = loc
        print(upload_place)

        if self.sandbox_exists(gsiloc+self.sbx_def['loc']+"/"+upload_name):
            self.delete_gsi_sandbox(gsiloc+self.sbx_def['loc']+"/"+upload_name)

        if self.sbxtarfile:
            upload = subprocess.Popen(['globus-url-copy', self.sbxtarfile, gsiloc +
                                       self.sbx_def['loc']+"/"+upload_name])
        upload.wait()
        print("Uploaded sandbox to "+gsiloc +
              self.sbx_def['loc']+"/"+upload_name)
        self.sbxloc = self.sbx_def['loc']+"/"+upload_name

    def sandbox_exists(self, sbxfile):
        file1 = subprocess.Popen(
            ['uberftp', '-ls', sbxfile], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output = file1.communicate()
        if output[0] != '' and output[1] == '':
            return True
        return False

    def delete_gsi_sandbox(self, sbxfile):
        deljob = subprocess.Popen(
            ['uberftp', '-rm', sbxfile], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        print("deleted old sandbox")
        return deljob.communicate()

    def zip_sbx(self, zipname=None):
        filename = zipname if zipname else self.sbx_def['name']+".tar"
        print(filename)
        tar = subprocess.call('tar -cf '+filename+' *', shell=True)
        print(tar)
        self.sbxtarfile = filename

    def cleanup(self):
        self.delete_sbx_folder()
        os.chdir(self.return_dir)

    def make_tokvar_dict(self):
        tokvardict = self.shell_vars
        with open('tokvar.cfg', 'w') as dumpfile:
            yaml.dump(tokvardict, dumpfile)

    def check_token(self):
        '''This function does the necessary linkage between the sandbox and token
           most importantly it saves the tokvar.cfg file in the sbx, but also checks
           if the token variables are all existing. If so, tokvar is created and put
           inside the SBX
        '''
        token_vars = self.tok_vars
        for key in self.shell_vars:
            if key in token_vars.keys():
                pass
            else:
                print(key+" missing")
        self.make_tokvar_dict()

    def get_result_loc(self):
        return (self.sbx_def['results']['UL_loc'] +
                "".join(self.sbx_def['results']['UL_pattern']))

    def build_sandbox(self, sbx_config):
        """A comprehensive function that builds a Sandbox from a configuration file and creates a
        sandbox tarfile.

        """
        self.parseconfig(sbx_config)
        self.create_sbx_folder()
        self.enter_sbx_folder()
        self.copy_base_scripts()
        self.load_git_scripts()
        self.make_tokvar_dict()
        self.zip_sbx()

    def upload_sandbox(self, upload_name=None):
        if self.authorized:
            self.upload_sbx(upload_name=upload_name)
        else: 
            warnings.warn("not authorized to uplad to gsiftp", RuntimeWarning)
            self.upload_ssh_sandbox(upload_name=upload_name)
        if self.sbx_def['remove_when_done'] == True:
            self.cleanup()

class UnauthorizedSandbox(Sandbox):
    def __init__(self, *args, **kw):
        super(UnauthorizedSandbox, self).__init__(*args, authorize=False, **kw)

